from openpyxl.styles import Font, Style, Alignment, Border, Side, PatternFill
from openpyxl.styles import borders, fills
from openpyxl.styles import Protection

from .utils import memoize


cell_fills = [
    PatternFill(
        fill_type=fills.FILL_SOLID,
        start_color='E6EFD7',
    ),
    PatternFill(
        fill_type=fills.FILL_SOLID,
        start_color='FFFFFF',
    ),
]
font = Font(name='Tahoma')
alignment = Alignment(
    vertical='center',
    shrink_to_fit=True,
    wrap_text=True,
    indent=1,
)
border = Side(border_style=borders.BORDER_THIN, color='B8D08A')
header_border = line_height = 13
title_style = Style(
    font=Font(name='Tahoma', size=20, bold=True)
)


@memoize
def column_header():
    return Style(
        font=Font(
            name='Thaoma',
            bold=True,
            color='FFFFFF',
            size=14,
        ),
        border=Border(
            left=Side(border_style=borders.BORDER_THIN, color='8BB048'),
            right=Side(border_style=borders.BORDER_THIN, color='8BB048'),
            top=Side(border_style=borders.BORDER_THIN, color='8BB048'),
            bottom=Side(border_style=borders.BORDER_THIN, color='8BB048'),
        ),
        alignment=Alignment(
            vertical='center',
            horizontal='center',
        ),
        fill=PatternFill(
            fill_type=fills.FILL_SOLID,
            start_color='8BB048',
        ),
    )


@memoize
def row_header(is_odd, is_last):
    return Style(
        font=font,
        alignment=Alignment(
            vertical='center',
            horizontal='center' if is_last else 'left',
            shrink_to_fit=True,
            wrap_text=True,
            indent=int(not is_last),
        ),
        border=Border(
            left=border,
            right=(Side(border_style=borders.BORDER_DOUBLE, color='8BB048')
                   if is_last else border),
            top=border,
            bottom=border,
        ),
        fill=cell_fills[int(bool(is_odd))],
        protection=Protection(),
    )


@memoize
def translation_cell(is_odd):
    return Style(
        font=font,
        alignment=Alignment(
            vertical='center',
            shrink_to_fit=True,
            wrap_text=True,
        ),
        border=Border(
            right=border,
            top=border,
            bottom=border,
        ),
        fill=cell_fills[int(bool(is_odd))],
        protection=Protection(locked=False),
    )


@memoize
def translation_cb_cell(is_odd):
    return Style(
        font=font,
        alignment=Alignment(
            vertical='center',
            horizontal='center',
        ),
        border=Border(
            left=border,
            top=border,
            bottom=border,
        ),
        fill=cell_fills[int(bool(is_odd))],
        protection=Protection(locked=False),
    )


comment_cell = translation_cell


@memoize
def occurrences_cell(is_odd):
    return Style(
        font=Font(name='Tahoma', size=8),
        alignment=Alignment(
            vertical='top',
            wrap_text=True,
        ),
        border=Border(
            right=border,
            top=border,
            bottom=border,
        ),
        fill=cell_fills[int(bool(is_odd))],
        protection=Protection(locked=False),
    )



from __future__ import print_function, unicode_literals

import os
import re
import math
import glob

import click
import polib
from openpyxl import load_workbook


def hr(char='\u2500', width=None, **kwargs):
    if width is None:
        width = click.get_terminal_size()[0]
    mult = int(math.ceil(width * 1.0 / len(char)))
    click.secho((char * mult)[:width], **kwargs)


def get_pofiles(locale_folders, language):
    for folder in locale_folders:
        rw_ro, folder = folder.split(':', 1)
        assert rw_ro in ['rw', 'ro']
        read_only = rw_ro == 'ro'
        po_file = os.path.join(folder, language, 'LC_MESSAGES', '*.po')
        for path in glob.glob(po_file):
            yield read_only, polib.pofile(path, wrapwidth=78)


class MsgStr(object):
    def __init__(self, entry, multiplicity=0):
        self.entry = entry
        self.multiplicity = multiplicity

    def get(self):
        if self.entry.msgid_plural:
            return self.entry.msgstr_plural[self.multiplicity]
        else:
            assert self.multiplicity == 0
            return self.entry.msgstr

    def normalize_whitespace(self, value):
        if value is None:
            return
        msgid = self.entry.msgid
        if self.entry.msgid_plural and self.multiplicity:
            msgid = self.entry.msgid_plural
        space_before = re.search(r'^\s*', msgid).group(0)
        space_after = re.search(r'\s*$', msgid).group(0)
        return ''.join([space_before, value.strip(), space_after])

    def set(self, value):
        if self.entry.msgid_plural:
            self.entry.msgstr_plural[self.multiplicity] = value
        else:
            self.entry.msgstr = value


@click.command()
@click.option('--skip', type=int, default=0)
@click.option('--ctx-col', '-c', type=int, default=2)
@click.option('--key-col', '-k', type=int, default=3)
@click.option('--trans-col', '-t', type=int, default=4)
@click.option('--sheet', '-s', type=int, default=1)
@click.option('--pretend/--no-pretend', '-p', default=False)
@click.option('--language', '-l')
@click.option('--add/--no-add', '-a')
@click.option('--hide-ok/--show-ok', default=False)
@click.argument('workbook', type=click.Path(exists=True))
@click.argument('locale_folders', nargs=-1)
def main(workbook, locale_folders, language, key_col, trans_col, sheet,
         pretend, hide_ok, ctx_col, skip, add):
    wb = load_workbook(workbook)
    ws = list(wb)[sheet-1]
    rows = iter(ws.iter_rows())

    for i in range(skip):
        next(rows)
    headers = next(rows)

    if key_col is None or trans_col is None:
        click.echo('The following columns are available:')
        for i, cell in enumerate(headers):
            if cell.value is not None:
                click.echo('{:2d}. {}'.format(i + 1, cell.value))

        if key_col is None:
            key_col = click.prompt('Which column contains the key?',
                                   type=int, default=1)

        if trans_col is None:
            trans_col = click.prompt('Which column contains the translation?',
                                     type=int)

    ctx_col -= 1
    key_col -= 1
    trans_col -= 1

    hr(fg='yellow')
    click.secho(' Selected sheet: {}'.format(ws.title))
    click.secho(' Key column:     {}'.format(headers[key_col].value))
    click.secho(' Value column:   {}'.format(headers[trans_col].value))
    click.secho(' Language:       {}'.format(language))
    hr(fg='yellow')

    pofiles = list(get_pofiles(locale_folders, language))

    for row in rows:
        ctx = row[ctx_col].value or None
        key = row[key_col].value
        plr = row[key_col + 1].value
        trans = row[trans_col].value

        # Check flag
        flag = int(row[trans_col - 1].value)
        fuzzy = (flag == 1)

        if key is None and trans is None:
            # This row can be safely ignored
            continue

        entry, msgstr, read_only = find_entry(pofiles, ctx, key, plr, trans)
        if entry:
            handle_entry(entry, msgstr, ctx, key, trans, fuzzy, hide_ok)
        elif plr == 's':
            # TODO: Support adding pluralized entries as well
            if add or click.confirm('Do you want to add it to the file?'):
                entry = polib.POEntry(
                    msgid=key or '',
                    msgstr=trans or '',
                    msgctx=ctx or None,
                )
                for read_only, po in pofiles:
                    if not read_only:
                        po.append(entry)
                        break

    if not pretend:
        for read_only, po in pofiles:
            if not read_only:
                po.save()


def find(pofile, st, by='msgid', include_obsolete_entries=False,
         msgctxt=False):
    if include_obsolete_entries:
        entries = pofile[:]
    else:
        entries = [e for e in pofile if not e.obsolete]
    for e in entries:
        if getattr(e, by).strip() == st:
            if msgctxt is not False and e.msgctxt != msgctxt:
                continue
            return e
    return None


def find_entry(pofiles, ctx, key, plr, trans):
    msgstr = None
    args_set = [
        {'include_obsolete_entries': False},
        {'include_obsolete_entries': True},
    ]

    for kwargs in args_set:
        if plr == 'P':
            kwargs['by'] = 'msgid_plural'
        for read_only, po in pofiles:
            entry = find(po, key, msgctxt=ctx, **kwargs)
            if entry is not None:
                msgstr = MsgStr(entry, plr == 'P')
                break
        if msgstr:
            break
    else:
        # TODO: How do we want to handle this?
        key = '{}:{}'.format(ctx, key) if ctx else key
        click.secho('The entry "{}" was not found.'.format(key), fg='red')
        return None, None, False

    if read_only:
        if trans != msgstr.get():
            key = '{}:{}'.format(ctx, key)
            click.secho('The entry "{}" has been translated as "{}" in "{}" '
                        'but the file is marked as read-only.'
                        .format(key, msgstr.get(), po.fpath), fg='red')
        return None, None, True

    return entry, msgstr, False


def handle_entry(entry, msgstr, ctx, key, trans, fuzzy, hide_ok):
    trans_norm = msgstr.normalize_whitespace(trans)

    if not trans or trans.strip() == msgstr.get().strip():
        # Already up to date
        if not hide_ok:
            click.secho('The entry "{}" is already up to date.'.format(key),
                        fg='green')
        if msgstr.get().strip():
            msgstr.set(msgstr.normalize_whitespace(msgstr.get()))
        else:
            msgstr.set('')
        if 'fuzzy' in entry.flags and not fuzzy:
            entry.flags.remove('fuzzy')
        entry.obsolete = False
        return

    if not entry.translated():
        # Entry is in the file, but has not been translated
        click.secho('The entry "{}" has not been translated yet.'
                    .format(key), fg='yellow')
        msgstr.set(trans_norm)
        if 'fuzzy' in entry.flags and not fuzzy:
            entry.flags.remove('fuzzy')
        # entry.obsolete = False
        return

    # Entry has been translated but needs update
    # TODO: Ask for confirmation
    hr()
    click.secho(str(ctx))
    click.secho('The entry "{}" was updated:'.format(key))
    click.secho(' OLD: "{}"'.format(msgstr.get()))
    click.secho(' NEW: "{}"'.format(trans_norm))
    msgstr.set(trans)
    hr()



import functools


def memoize(obj):
    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        key = (args, tuple(kwargs.items()))
        if key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]
    return memoizer



import os
import glob

import click

import polib


class LocaleFolderParamType(click.Path):
    name = 'locale-folder'

    def __init__(self):
        super(LocaleFolderParamType, self).__init__(
            exists=True, file_okay=False, dir_okay=True, writable=True,
            readable=True, resolve_path=True)

    def convert(self, value, param, ctx):
        path = super(LocaleFolderParamType, self).convert(value, param, ctx)
        return LocaleFolder(path)


class LocaleFolder(object):
    def __init__(self, path):
        self.path = path

    def locales(self):
        locales = os.path.join(self.path, '*', 'LC_MESSAGES')

        for path in glob.glob(locales):
            locale = os.path.basename(os.path.dirname(path))
            yield locale

    def pofiles(self, *args, **kwargs):
        po_files = os.path.join(self.path, '*', 'LC_MESSAGES', '*.po')
        for path in glob.glob(po_files):
            locale = os.path.basename(os.path.dirname(os.path.dirname(path)))
            yield locale, polib.pofile(path, *args, **kwargs)



from __future__ import print_function, unicode_literals

import collections
from textwrap import wrap

import click

from openpyxl import Workbook
from openpyxl.cell import get_column_letter

from i18n_utils import styles, params


@click.command()
@click.option('--locale', '-l', 'locales', multiple=True)
@click.option('--key-locale', '-k', default='en')
@click.option('--title', '-t', default='My project')
@click.argument('out', type=click.Path(exists=False, dir_okay=False,
                                       writable=True))
@click.argument('folders', nargs=-1, type=params.LocaleFolderParamType())
def main(locales, key_locale, title, folders, out):
    wb = Workbook()

    if not locales:
        locales = list(folders[0].locales())

    add_translations_wb(wb.active, folders, locales, key_locale, title)
    wb.save(out)


def coord(col, row):
    if str(col).isdigit():
        col = get_column_letter(int(col))
    return '{}{}'.format(col, row)


def num_lines(s, chars=50):
    lines = s.split('\n')
    return sum(len(wrap(l, chars)) for l in lines)


def add_row(ws, left, context, key, key_idx, entries, plural=False):
    if plural:
        def get_str(entry):
            return entry.msgstr_plural[1].strip()
    else:
        def get_str(entry):
            if entry.msgid_plural:
                return entry.msgstr_plural[0].strip()
            else:
                return entry.msgstr.strip()

    pluralized = plural
    if not pluralized:
        for e in entries:
            if e:
                pluralized = bool(e.msgid_plural)
                break

    ws.append([''] * (len(entries) * 2 + 5 + left))
    row_idx = len(ws.rows)
    row = ws.rows[-1][left:]

    occurrences = sum((entry.occurrences for entry in entries if entry), [])
    occurrences_dict = collections.defaultdict(lambda: set())
    for path, line in occurrences:
        occurrences_dict[path].add(line)

    def fmt(o, lines):
        return '{}:{}'.format(o, ','.join(sorted(lines)))

    occurrences = '\n'.join(fmt(*o) for o in sorted(occurrences_dict.items()))

    # Set row height
    content = [context or '', key] + [get_str(e) for e in entries if e]
    lines = max(num_lines(v) for v in content)
    height = (lines + 1) * styles.line_height
    height = max(height, 11 * num_lines(occurrences, 200) + 3)
    ws.row_dimensions[row_idx].height = height

    # Add context cell
    ctx_cell = row[0]
    ctx_cell.value = context
    ctx_cell.style = styles.row_header(is_odd=row_idx % 2, is_last=False)

    # Add key cell
    key_cell = row[1]
    key_cell.value = key
    key_cell.style = styles.row_header(is_odd=row_idx % 2, is_last=False)

    # Add plural cell
    plr_cell = row[2]
    plr_cell.value = 'P' if plural else 'p' if pluralized else 's'
    plr_cell.style = styles.row_header(is_odd=row_idx % 2, is_last=True)

    # Add translations cells
    row = row[3:]
    for i, entry in enumerate(entries):
        cb_cell = row[i * 2]
        msgstr = get_str(entry) if entry else ''
        if (entry and msgstr and 'fuzzy' not in entry.flags
                and not entry.obsolete):
            msg = msgstr
            cb_cell.value = 2
        elif entry and msgstr and 'fuzzy' in entry.flags:
            msg = msgstr
            cb_cell.value = 1
        else:
            if i == key_idx:
                msg = key
                cb_cell.value = 2
            else:
                msg = ''
                cb_cell.value = 0

        cb_cell.style = styles.translation_cb_cell(is_odd=row_idx % 2)

        trans_cell = row[i * 2 + 1]
        trans_cell.value = msg
        trans_cell.style = styles.translation_cell(is_odd=row_idx % 2)

    # Add comment
    cmt_cell = row[-2]
    cmt_cell.style = styles.comment_cell(is_odd=row_idx % 2)

    # Occurrences
    occ_cell = row[-1]
    occ_cell.style = styles.occurrences_cell(is_odd=row_idx % 2)
    occ_cell.value = occurrences


def add_translations_wb(ws, folders, locales, key_locale, title):
    top = 4
    left = 1

    ws.title = 'Translations'

    # Parse the PO files
    entries = collections.defaultdict(lambda: [None] * len(locales))

    for folder in folders:
        po_files = [(locale, po) for locale, po in folder.pofiles()
                    if locale in locales]

        for locale, po in po_files:
            click.secho('Parsing {}'.format(po.fpath), fg='yellow')
            idx = locales.index(locale)
            for entry in po:
                if entry.obsolete:
                    continue
                if locale == key_locale and not entry.msgstr:
                    continue

                key = entry.msgctxt, entry.msgid.strip(), False

                try:
                    current_entry = entries[key][idx]
                except (KeyError, IndexError):
                    pass
                else:
                    if current_entry:
                        if current_entry.msgid_plural:
                            current_entry.msgstr_plural[0].strip()
                        else:
                            msgstr = current_entry.msgstr.strip()

                        if msgstr:
                            continue

                entries[key][idx] = entry

                # Handle plurals
                if not entry.msgid_plural:
                    continue

                key = entry.msgctxt, entry.msgid_plural.strip(), True

                try:
                    current_entry = entries[key][idx]
                except (KeyError, IndexError):
                    pass
                else:
                    if current_entry:
                        if current_entry.msgstr_plural[1].strip():
                            continue

                entries[key][idx] = entry

    entries = sorted(entries.items(), key=lambda e: (e[0][0], e[0][1].lower()))
    key_idx = locales.index(key_locale)

    # Add the translations
    # ws = wb.create_sheet()
    ws.sheet_view.showGridLines = False

    # Sheet title ############################################################
    header = ws.cell(coord(left + 1, top - 2))
    header.value = 'Translations for "{}"'.format(title)
    header.style = styles.title_style

    # Headers ################################################################
    ctx_cell = ws.cell(coord(left + 1, top))
    ctx_cell.style = styles.column_header()
    ctx_cell.value = 'Context'

    ws.merge_cells('{}:{}'.format(coord(left + 2, top), coord(left + 3, top)))
    key_cell = ws.cell(coord(left + 2, top))
    key_cell.style = styles.column_header()
    key_cell.value = 'Key ({} entries)'.format(len(entries))
    ws.cell(coord(left + 3, top)).style = styles.column_header()

    for i, l in enumerate(locales):
        from_cell = coord(i * 2 + 4 + left, top)
        to_cell = coord(i * 2 + 4 + left + 1, top)
        ws.merge_cells('{}:{}'.format(from_cell, to_cell))
        locale_header_text = (
            '=CONCATENATE('
            '    "{name} (",'
            '    ROUND(COUNTIF({start}:{end}, "=2")/{count} *100, 0),'
            '    "% translated)"'
            ')'
        )
        cell = ws.cell(from_cell)
        cell.style = styles.column_header()
        cell.value = locale_header_text.format(
            name=l.upper(),
            start=coord(i * 2 + 4 + left, top + 1),
            end=coord(i * 2 + 4 + left, len(entries) + top + 1),
            count=len(entries)
        )
        ws.cell(to_cell).style = styles.column_header()

    cmt_cell = ws.cell(coord(len(locales) * 2 + 4 + left, top))
    cmt_cell.style = styles.column_header()
    cmt_cell.value = 'Comment'

    occ_cell = ws.cell(coord(len(locales) * 2 + 5 + left, top))
    occ_cell.style = styles.column_header()
    occ_cell.value = 'Occurrences'

    # Add translations #######################################################
    for i, ((ctx, key, plural), translations) in enumerate(entries):
        add_row(ws, left, ctx, key, key_idx, translations, plural=plural)

    # Column widths ##########################################################

    # Left offset
    for i in range(left):
        ws.column_dimensions[get_column_letter(i + 1)].width = 5

    # Context and key
    ws.column_dimensions[get_column_letter(left + 1)].width = 30
    ws.column_dimensions[get_column_letter(left + 2)].width = 50
    ws.column_dimensions[get_column_letter(left + 3)].width = 5

    # Translations
    for i in range(len(locales)):
        col = left + 4 + i * 2
        ws.column_dimensions[get_column_letter(col)].width = 5
        ws.column_dimensions[get_column_letter(col + 1)].width = 50

    # Comment
    col = left + 4 + len(locales) * 2
    ws.column_dimensions[get_column_letter(col)].width = 100

    # Occurrences
    col = left + 5 + len(locales) * 2
    ws.column_dimensions[get_column_letter(col)].width = 80

    # Row heights ############################################################
    for i in range(top):
        ws.row_dimensions[i + 1].height = 25
    ws.row_dimensions[top].height = 40

    # Conditional formatting #################################################
    for i in range(len(locales)):
        from_cell = coord(left + 4 + i * 2, top + 1)
        to_cell = coord(left + 4 + i * 2, top + 1 + len(entries))
        ws.conditional_formatting.add(
            '{}:{}'.format(from_cell, to_cell),
            {
                'iconSet': {'cfvo': [{'type': 'num', 'val': '0'},
                                     {'type': 'num', 'val': '1'},
                                     {'type': 'num', 'val': '2'}],
                            'iconSet': '3Flags',
                            'showValue': '0'},
                'priority': 1,
                'type': 'iconSet'
            }
        )

    # Freeze and protect #####################################################
    ws.freeze_panes = ws[coord(left + 4, top + 1)]
    ws.protection.enable()



import click

from i18n_utils import params, normalization


@click.command()
@click.option('--wrap-width', '-w', 'wrapwidth',
              default=normalization.DEFAULT_WRAPPING_WIDTH)
@click.argument('folders', nargs=-1, type=params.LocaleFolderParamType())
def main(folders, wrapwidth):
    """Normalizes the PO files in the given locale folders.

    Useful to remove unecessary noise from source commits. The normalization
    process applies the following operations:

    * The lines are rewrapped to the given wrapwidth
    * Sort the entries in the file by their `msgstr` (obsoleted entries are
      sorted in line with current entries and not grouped together in the end)
    * Each occurrence is put on a line by itself, even if the wrapwidth would
      allow for more to be put on a single line
    """
    for f in folders:
        for locale, pofile in f.pofiles(
                wrapwidth=wrapwidth, klass=normalization.NormalizedPOFile):
            pofile.save()



import six

import polib


DEFAULT_WRAPPING_WIDTH = 78


def one_occurrence_per_line(entry_str):
    entry_lines = []

    for line in entry_str.split('\n'):
        if line.startswith('#: '):
            line = line[3:]
            for occurrence in line.split(' '):
                entry_lines.append('#: {}'.format(occurrence))
        else:
            entry_lines.append(line)

    return u'\n'.join(entry_lines)


def entry_to_unicode(entry, wrapwidth):
    return one_occurrence_per_line(entry.__unicode__(wrapwidth))


class NormalizedPOFile(polib.POFile):
    def __unicode__(self):
        """
        Returns the normalized unicode representation of the file.
        """
        meta, headers = '', self.header.split('\n')
        for header in headers:
            if header[:1] in [',', ':']:
                meta += u'#%s\n' % header
            else:
                meta += u'# %s\n' % header

        ret = []
        entries = [self.metadata_as_entry()]
        entries += sorted(self, key=lambda e: e.msgid)
        for entry in entries:
            ret.append(entry_to_unicode(entry, self.wrapwidth))
        ret = u'\n'.join(ret)
        assert isinstance(ret, six.text_type)
        return meta + ret
